Unit Form_ViewFull;

Interface

Uses
    Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
    Dialogs, StdCtrls, ExtCtrls, GR32, GR32_Image;

Type
    TViewFull = Class(TForm)
        Panel1: TPanel;
        cbTopLayer: TComboBox;
        Label1: TLabel;
        View: TImgView32;
        cbScale: TComboBox;
        Label2: TLabel;
        cSelectedOnly: TCheckBox;
        Procedure FormShow(Sender: TObject);
        Procedure FormCreate(Sender: TObject);
        Procedure cbTopLayerCloseUp(Sender: TObject);
        Procedure cbScaleChange(Sender: TObject);
        Procedure cSelectedOnlyClick(Sender: TObject);
    Private
    Public
        maxz, minz, difz: Integer;
        Procedure Load;
        Procedure RenderLayer(LayerNo: Integer);
        Procedure RenderLayers;
        Procedure ReCalcMaxZ;
    End;

Var
    ViewFull             : TViewFull;

Implementation

Uses Unit_GlobalVars;

{$R *.dfm}

Procedure TViewFull.FormShow(Sender: TObject);
Begin
    Load;
End;

Procedure TViewFull.FormCreate(Sender: TObject);
Begin
    cSelectedOnly.Checked := False;
    View.Bitmap.SetSize(2048, 2048);
End;

Procedure TViewFull.Load;
Var
    k                    : Integer;
Begin
    View.Bitmap.Clear;

    cbTopLayer.Clear;

    For k := 0 To _LastLayer Do
        cbTopLayer.Items.Add('L' + inttostr(k));
    cbTopLayer.ItemIndex := _LastLayer;

    If Not GeoReady Then
    Begin
        ShowMessage('Geo not loaded');
        Exit;
    End;

    RenderLayers;
End;

Procedure TViewFull.RenderLayer(LayerNo: Integer);
Var
    x, y                 : Integer;
    z                    : Integer;
    c                    : TColor32;
Begin
    If LayerNo > __MaxLayer Then exit;
    If LayerNo <0 Then exit;

    If LayerNo = 0 Then
    Begin
        maxz := -MaxInt;
        minz := MaxInt;
        For y := 0 To 2047 Do
            For x := 0 To 2047 Do
            Begin
                z := GeoLayer[0][x, y].z;
                If z > maxz Then maxz := z;
                If z < minz Then minz := z;
            End;
        difz := abs(maxz - minz);
        If difz = 0 Then difz := 1;
    End;

    For y := 0 To 2047 Do
        For x := 0 To 2047 Do
        Begin
            If GeoLayer[LayerNo][x, y].m <> $FF Then
            Begin
                z := GeoLayer[LayerNo][x, y].z - minz;
                if colormodel=1 then
                begin
                  //c := z * 255*255 Div difz;
                  c := ColorModelBmp.PixelS[z * (ColorModelBmp.Width-1) Div difz,0];
                end
                else
                begin
                  z := z * 255 Div difz;
                  c := Gray32(z);
                  If GeoLayer[LayerNo][x, y].z < WaterLevel Then c := c Or $40;
                end;

                //if GeoLayer[LayerNo][x, y].z<(GeoHead.wZ) then c := c or $40;
                //If GeoLayer[LayerNo][x, y].z < WaterLevel Then c := c Or $40;

                If GeoLayer[LayerNo][x, y].m And $F <> $F Then
                    c := c Or $800000; //non-drop
                If (cSelectedOnly.Checked = False) And (LayerNo > 0) Then
                Begin
                    //add alpha to 1+ layers
                    c := SetAlpha(c,80);
                End
                else
                begin
                   c := SetAlpha(c,255);
                end;
                View.Bitmap.PixelS[y, x] := c;
            End;
        End;
End;

Procedure TViewFull.RenderLayers;
Var
    i                    : Integer;
Begin
    Screen.Cursor := crHourGlass;
    View.Bitmap.Clear;
    If cSelectedOnly.Checked = False Then
    Begin
        For i := 0 To cbTopLayer.ItemIndex Do
        Begin
            RenderLayer(i);
        End;
    End
    Else
    Begin
        RenderLayer(cbTopLayer.ItemIndex);
    End;
    View.Bitmap.Changed;
    View.Invalidate;
    Screen.Cursor := crDefault;
End;

Procedure TViewFull.cbTopLayerCloseUp(Sender: TObject);
Begin
    RenderLayers;
End;

Procedure TViewFull.cbScaleChange(Sender: TObject);
Begin
    Case cbScale.ItemIndex Of
        0: View.Scale := 0.5;
        2: View.Scale := 1.5;
        3: View.Scale := 2.0;
    Else
        View.Scale := 1;
    End;
End;

Procedure TViewFull.cSelectedOnlyClick(Sender: TObject);
Begin
    RenderLayers;
End;

Procedure TViewFull.ReCalcMaxZ;
var
  x,y,z: integer;
Begin
    maxz := -MaxInt;
    minz := MaxInt;
    For y := 0 To 2047 Do
        For x := 0 To 2047 Do
        Begin
            z := GeoLayer[0][x, y].z;
            If z > maxz Then maxz := z;
            If z < minz Then minz := z;
        End;
    difz := abs(maxz - minz);
    If difz = 0 Then difz := 1;
End;

End.

