Unit Unit_GlobalVars;

Interface

Uses Windows, Classes, Dialogs, SysUtils, GR32, GR32_Image, GR32_Convolution,
    GR32_Resamplers, messages, Forms;


Const
    WaterLevel           = -4100;
    WorkSpaceID          = 'HDGEOWORKSPACE';

Var
    ColorModel           : Integer = 1;
    ColorModelBmp        : TBitmap32;
    WorldAddX            : Integer = 8;
    WorldAddY            : Integer = 8;


    //    MapInfoBase          = $01245228;
    PlayerInfoBaseStat   : DWord = $001234EC;
    PlayerInfoBasePtr    : DWord = $0012BFC8;
    //PlayerInfoBasePtr    = $0012F42C;
    PlayerInfoBaseAdd    : DWord = $1B0; //$1B0;// $184;

    //08DCA984 08DCA9B0 08DCAB20 08E00954

    //$0011AF28; //$0011ABE8; //$0011B268;
    //$0011BF7C; 1234EC 12382C
    //00123B6C 001241EC 0012F3C0 0012F42C 001241EC

Type
    TWorkSpaceHeader = Packed Record
        ID: String[20];
        MX, MY: Byte;
        ShowWater: boolean;
        FirstX, FirstY: Integer;
        SelectedX, SelectedY: Integer;
        LastLayer: Integer;
        SelectedLayer: Integer;
        SelectedOnly: boolean;
        OptimaView: boolean;
        SelectedLeft, SelectedTop, SelectedRight, SelectedBottom: Integer;
        DrawZ: Integer;
        Memo: String[255];
        Reserved: Array[1..50] Of Integer;
    End;

    TGeoHeader = Packed Record
        MX: Byte;
        MY: Byte;
        w0080: Word;
        w0010: Word;
        dw1: DWord;
        dw2: DWord;
        dw3: DWord;
    End;

    TLayerCell0 = Packed Record
        z: SmallInt;
        m: Byte;
    End;

{    TLayerCell = Packed Record
        x, y, z: SmallInt;
        m: Byte;
    End;}

    TGeoLayer = Array[0..2047, 0..2047] Of TLayerCell0;

    TCellType = (ctNone, ct0000, ct4000, ct4100);
    TGeoError = (geOk, geFileNotFound, geInvalidFormat, geTooManyLayers);

var
    __MaxLayer: integer            = 32;
//const
//      _MaxLayer            = 32;


Var
    GeoHead              : TGeoHeader;
    GeoLayer             : Array Of TGeoLayer;
    //GeoLayer             : Array [0.._MaxLayer] Of TGeoLayer;
    //CellTypes            : Array[0..255, 0..255] Of TCellType;
    _LastLayer           : Integer;
    GeoReady             : boolean = False;

    GeoTemp              : TGeoLayer;
    GeoTempX1            : Integer = 0;
    GeoTempY1            : Integer = 0;
    GeoTempX2            : Integer = 0;
    GeoTempY2            : Integer = 0;
    GeoTempTrue          : boolean = False;
    GeoTempLayer         : Integer;
    GeoTempMapX          : Byte;
    GeoTempMapY          : Byte;


Type
    TPlayerInfo = Record
        x, y, z: single;
    End;
    TMapInfo = Record
        x, y: DWord;
    End;

Const
    FileBufferID         = 'HDEDITOR_BUFFER';
Type
    TFileBufferHeader = Record
        ID: String[30];
        Info: String[255];
        x1, y1, x2, y2: Integer;
        mx, my, l: Integer;
    End;

Var
    _ClientPlayerInfo    : TPlayerInfo;
    ClientMapInfo        : TMapInfo;

    ClientPlayerInfoX    : Integer;
    ClientPlayerInfoY    : Integer;
    ClientPlayerInfoZ    : Integer;

Procedure ClearLayers;
Function GeoErrorToStr(err: TGeoError): String;
Function LoadGeoMap(FName: String): TGeoError;
//Function LoadGeoMap09(FName: String): TGeoError;
Function LoadGeoMap256(MapX, MapY: Integer; b256: TBitmap32; minz, maxz,
    difz: smallint): TGeoError;

Procedure SaveGeoMapTurbo(FName: String);

Function _GetWorldX(MapX, CellX: Integer; SubX: Integer = 0): Integer;
Function _GetWorldY(MapY, CellY: Integer; SubY: Integer = 0): Integer;
Procedure _WorldToXY(wx, wy: Integer; Var CellX, CellY: Integer);

Function GetLineageWnd: HWnd;
Procedure PostKeysToLineAge(keys: String; addEnter: boolean = True);
Function ScanClientMemory(fromstat: boolean = True): boolean;

Function isGeoCoord(x: Integer; y: Integer = 0): boolean;

Procedure FillRecordZ(l, _x, _y, mode: Integer); //mode 0=room 1=rock 2=hole

Procedure GeoLayerToTemp(l, x1, y1, x2, y2: Integer);
Procedure GeoTempToLayer(l, tox, toy: Integer; writeZero: boolean = False);

Procedure SaveTempToFile(FName: String);
Procedure LoadTempFromFile(FName: String);
Procedure c_teleport(x, y: Integer; z: Integer = -15001);

Procedure _Normalize0; //normalize layer0  //[x,y].m &= 1F

Implementation

Uses ArraySort, Form_Main, Form_CellEditor, Form_Progress;

Procedure _Normalize0; //normalize layer0  //[x,y].m &= 1F
Var
    x, y                 : Integer;
Begin
    For y := 0 To 2047 Do
        For x := 0 To 2047 Do
            GeoLayer[0][x, y].m := GeoLayer[0][x, y].m And $1F;
End;

Procedure c_teleport(x, y: Integer; z: Integer = -15001);
Var
    s                    : String;
Begin
    s := '///c_teleport x=' + inttostr(x) + ' y=' + inttostr(y);
    If z > -15001 Then s := s + ' z=' + inttostr(z);
    PostKeysToLineAge(s);
End;

Procedure SaveTempToFile(FName: String);
Var
    s                    : String;
    head                 : TFileBufferHeader;
    f                    : File;
    x, y                 : Integer;
Begin
    If Not GeoTempTrue Then exit;
    s := 'From Map: ' + inttostr(GeoTempMapX) + '_' + inttostr(GeoTempMapY);
    s := s + #13'From Layer: ' + inttostr(GeoTempLayer);
    s := s + #13'From Selection: ' + inttostr(GeoTempX1 Div 8) + ' ' +
        inttostr(GeoTempY1 Div 8) + ' (' +
        inttostr(GeoTempX1 Mod 8) + ' ' + inttostr(GeoTempY1 Mod 8) + ') - ' +
        inttostr(GeoTempX2 Div 8) + ' ' + inttostr(GeoTempY2 Div 8) + ' (' +
        inttostr(GeoTempX2 Mod 8) + ' ' + inttostr(GeoTempY2 Mod 8) + ')';
    If Application.MessageBox(PChar(s), 'SAVE BUFFER TO FILE?', MB_YESNO) <>
        IDYES Then exit;
    s := s + #13'Saved: ' + DateTimeToStr(Now);

    FillChar(head, sizeof(head), 0);
    head.ID := FileBufferID;
    head.Info := s;
    head.x1 := GeoTempX1;
    head.y1 := GeoTempY1;
    head.x2 := GeoTempX2;
    head.y2 := GeoTempY2;
    head.mx := GeoTempMapX;
    head.my := GeoTempMapY;
    head.l := GeoTempLayer;
    FileMode := 2;
    AssignFile(f, FName);
    Rewrite(f, 1);
    BlockWrite(f, head, sizeof(head));
    For y := GeoTempY1 To GeoTempY2 Do
    Begin
        For x := GeoTempX1 To GeoTempX2 Do
        Begin
            BlockWrite(f, GeoTemp[x, y], sizeof(TLayerCell0));
        End;
    End;
    CloseFile(f);
End;

Procedure LoadTempFromFile(FName: String);
Var
    head                 : TFileBufferHeader;
    s                    : String;
    f                    : File;
    x, y                 : Integer;
Begin
    FileMode := 0;
    AssignFile(f, FName);
    Reset(f, 1);
    BlockRead(f, head, sizeof(head));
    If head.ID <> FileBufferID Then
    Begin
        CloseFile(f);
        ShowMessage('ERROR: Invalid GEB Format');
        exit;
    End;

    s := head.Info;
    If Application.MessageBox(PChar(s), 'READ BUFFER FROM FILE?',
        MB_YESNO) <> IDYES Then
    Begin
        CloseFile(f);
        exit;
    End;

    GeoTempX1 := head.x1;
    GeoTempY1 := head.y1;
    GeoTempX2 := head.x2;
    GeoTempY2 := head.y2;

    For y := GeoTempY1 To GeoTempY2 Do
    Begin
        For x := GeoTempX1 To GeoTempX2 Do
        Begin
            BlockRead(f, GeoTemp[x, y], sizeof(TLayerCell0));
        End;
    End;
    CloseFile(f);

    GeoTempTrue := True;

End;

Procedure GeoLayerToTemp(l, x1, y1, x2, y2: Integer);
Var
    x, y                 : Integer;
Begin
    If x1 > x2 Then exit;
    If y1 > y2 Then exit;
    If Not isGeoCoord(x1, y1) Then exit;
    If Not isGeoCoord(x2, y2) Then exit;
    GeoTempX1 := x1;
    GeoTempY1 := y1;
    GeoTempX2 := x2;
    GeoTempY2 := y2;
    GeoTempLayer := l;
    GeoTempMapX := GeoHead.MX;
    GeoTempMapY := GeoHead.MY;
    For y := y1 To y2 Do
    Begin
        For x := x1 To x2 Do
        Begin
            GeoTemp[y, x].z := GeoLayer[l][y, x].z;
            GeoTemp[y, x].m := GeoLayer[l][y, x].m;
        End;
    End;
    GeoTempTrue := True;
End;

Procedure GeoTempToLayer(l, tox, toy: Integer; writeZero: boolean = False);
Var
    dx, dy, x, y, x1, y1, x2, y2: Integer;

Begin
    If Not GeoTempTrue Then exit;
    If Not isGeoCoord(tox, toy) Then exit;
    x1 := GeoTempX1;
    y1 := GeoTempY1;
    x2 := GeoTempX2;
    y2 := GeoTempY2;
    If x1 > x2 Then exit;
    If y1 > y2 Then exit;
    If Not isGeoCoord(x1, y1) Then exit;
    If Not isGeoCoord(x2, y2) Then exit;
    dy := 0;
    For y := y1 To y2 Do
    Begin
        dx := 0;
        For x := x1 To x2 Do
        Begin
            If Not writeZero Then
            Begin
                If GeoTemp[y, x].m <> $FF Then
                Begin
                    GeoLayer[l][toy + dy, tox + dx].z := GeoTemp[y, x].z;
                    GeoLayer[l][toy + dy, tox + dx].m := GeoTemp[y, x].m;
                End;
            End
            Else
            Begin
                GeoLayer[l][toy + dy, tox + dx].z := GeoTemp[y, x].z;
                GeoLayer[l][toy + dy, tox + dx].m := GeoTemp[y, x].m;
            End;
            inc(dx);
        End;
        inc(dy);
    End;
    GeoTempTrue := True;
End;


Function isGeoCoord(x: Integer; y: Integer = 0): boolean;
Begin
    If (x >= 0) And (x < 2048) And (y >= 0) And (y < 2048) Then result := True
    Else result := False;
End;

Procedure FillRecordZ(l, _x, _y, mode: Integer); //mode 0=room 1=rock 2=hole
Const
    difz                 = 100;
Var
    m, i                 : SmallInt;
    x, y                 : Integer;
    x1, x2               : Integer; //left/right X
    z1, z2               : SmallInt; //left/right Z
    z, zincr             : real; //Z increment
    miny, maxy, minx, maxx: Integer;
Begin

    If l < 0 Then exit;
    If l > __MaxLayer Then exit;
    If _x < 0 Then exit;
    If _y < 0 Then exit;
    If _x > 2047 Then exit;
    If _y > 2047 Then exit;
    If (mode = 2) And (l = 0) Then exit; //cand make hole on layer[0]

    m := 0;
    //horizontal pass
    //find minY
    miny := _y;
    While miny > 0 Do
    Begin
        m := GeoLayer[l][miny, _x].m;
        If m And $10 > 0 Then Break;
        dec(miny);
    End;
    If m And $10 = 0 Then exit; //RecordZ not found

    //find maxY
    maxy := _y;
    While maxy < 2047 Do
    Begin
        m := GeoLayer[l][maxy, _x].m;
        If m And $10 > 0 Then Break;
        inc(maxy);
    End;
    If m And $10 = 0 Then exit; //RecordZ not found

    y := miny;
    minx := 2046;
    maxx := 1;

    //for each line...
    While y <= maxy Do
    Begin
        //search left and right X
        x1 := _x;
        While x1 > 0 Do
        Begin
            m := GeoLayer[l][y, x1].m;
            If m And $10 > 0 Then
                If GeoLayer[l][y, x1 - 1].m And $10 = 0 Then Break;
            dec(x1);
        End;
        If m And $10 = 0 Then exit; //RecordZ not found
        x2 := _x;
        While x2 < 2047 Do
        Begin
            m := GeoLayer[l][y, x2].m;
            If m And $10 > 0 Then
                If GeoLayer[l][y, x2 + 1].m And $10 = 0 Then Break;
            inc(x2);
        End;
        If x1 < minx Then minx := x1;
        If x2 > maxx Then maxx := x2;

        If m And $10 = 0 Then exit; //RecordZ not found

        If x1 = x2 Then x2 := x1 + 1; //oops

        //do calcs
        z1 := GeoLayer[l][y, x1].z;
        z2 := GeoLayer[l][y, x2].z;
        z := z1;
        zincr := (z2 - z1) / (x2 - x1);

        If mode = 0 Then
        Begin
            //draw Z
            For i := x1 To x2 Do
            Begin
                GeoLayer[l][y, i].m := GeoLayer[l][y, i].m Or $10;
                If (i = x1) Or (i = x2) Or (y = miny) Or (y = maxy) Then
                    GeoLayer[l][y, i].z := round(z) + difz + 1
                Else
                    GeoLayer[l][y, i].z := round(z);
                z := z + zincr;
            End;
        End;
        If mode = 1 Then
        Begin
            //draw Z
            For i := x1 To x2 Do
            Begin
                GeoLayer[l][y, i].m := GeoLayer[l][y, i].m Or $10;
                If (i = x1) Or (i = x2) Or (y = miny) Or (y = maxy) Then
                    GeoLayer[l][y, i].z := round(z)
                Else
                    GeoLayer[l][y, i].z := round(z) + difz + 1;
                z := z + zincr;
            End;
        End;
        inc(y);
    End;

    //tune up movement restrictions
{    For y := miny To maxy Do
        For x := minx To maxx Do
        Begin
            If GeoLayer[l][y, x].m And $10 > 0 Then
            Begin
                If GeoLayer[l][y + 1, x].z - GeoLayer[l][y, x].z > difz Then
                    GeoLayer[l][y, x].m := GeoLayer[l][y, x].m And (11 Or $10);
                If GeoLayer[l][y - 1, x].z - GeoLayer[l][y, x].z > difz Then
                    GeoLayer[l][y, x].m := GeoLayer[l][y, x].m And (7 Or $10);
                If GeoLayer[l][y, x + 1].z - GeoLayer[l][y, x].z > difz Then
                    GeoLayer[l][y, x].m := GeoLayer[l][y, x].m And (14 Or $10);
                If GeoLayer[l][y, x - 1].z - GeoLayer[l][y, x].z > difz Then
                    GeoLayer[l][y, x].m := GeoLayer[l][y, x].m And (13 Or $10);

                //restore down Z
//                If GeoLayer[l][y, x].m And $30 > 0 Then
//                begin
//                    GeoLayer[l][y, x].m := GeoLayer[l][y, x].m and $1F;
//                    GeoLayer[l][y, x].z := GeoLayer[l][y, x].z + difz + 1;
//                end;

            End;
        End;}

End;


Function MaxPtLayer(x, y: Integer): Integer;
Var
    l                    : Integer;
Begin
    result := 0;
    For l := 1 To _LastLayer Do
        If GeoLayer[l][x, y].m <> $FF Then result := result + 1;
End;

Function GetLineageWnd: HWnd;
Begin
    result := FindWindow('l2UnrealWWindowsViewportWindow', 'Lineage II');
End;

Procedure PostKeysToLineAge(keys: String; addEnter: boolean);
Var
    h                    : HWnd;
    i                    : Integer;
Begin
    h := GetLineageWnd;
    If h = 0 Then exit;
    SetForegroundWindow(h);
    For i := 1 To length(keys) Do
    Begin
        SendMessage(h, WM_CHAR, word(keys[i]), 0);
    End;
    If addEnter Then
    Begin
        SendMessage(h, WM_KEYDOWN, VK_RETURN, 0);
        SendMessage(h, WM_KEYUP, VK_RETURN, 0);
    End;
End;


Function ScanClientMemory(fromstat: boolean = True): boolean;
Var
    h                    : HWnd;
    ph, pid              : DWord;
    Address              : Pointer;
    AddrFrom             : Pointer;
    BytesRead            : Cardinal;
    r                    : boolean;
    x, y                 : Integer;
Begin
    result := False;

    FillChar(_ClientPlayerInfo, sizeof(_ClientPlayerInfo), 0);

    h := GetLineageWnd;
    If h = 0 Then exit;
    GetWindowThreadProcessID(h, @pid);

    ph := OpenProcess(PROCESS_VM_READ, False, pid);
    If ph = 0 Then exit;

    If Not fromstat Then
    Begin
        Address := Pointer(PlayerInfoBasePtr);
        ReadProcessMemory(ph, Address, @AddrFrom, sizeof(Pointer),
            BytesRead);
        AddrFrom := Pointer(Cardinal(AddrFrom) + PlayerInfoBaseAdd);
        ReadProcessMemory(ph, AddrFrom, @_ClientPlayerInfo, sizeof(_ClientPlayerInfo),
            BytesRead);
        //        Address := Pointer(MapInfoBase);
        //        ReadProcessMemory(ph, Address, @ClientMapInfo, sizeof(ClientMapInfo),   BytesRead);
    End
    Else
    Begin
        Address := Pointer(PlayerInfoBaseStat);
        ReadProcessMemory(ph, Address, @_ClientPlayerInfo, sizeof(_ClientPlayerInfo),
            BytesRead);
    End;
    CloseHandle(ph);

    Try
        ClientPlayerInfoX := trunc(_ClientPlayerInfo.x);
        ClientPlayerInfoY := trunc(_ClientPlayerInfo.y);
        ClientPlayerInfoZ := trunc(_ClientPlayerInfo.z);
        r := (ClientPlayerInfoX <> 0) And (ClientPlayerInfoY <> 0) And
            (ClientPlayerInfoZ <> 0);
    Except
        r := False;
    End;

    If r Then
    Begin
        _WorldToXY(ClientPlayerInfoX, ClientPlayerInfoY, x, y);
        If (x < 0) Or (y < 0) Then r := False;
    End;

    If Not r Then
    Begin
        ClientPlayerInfoX := 0;
        ClientPlayerInfoY := 0;
        ClientPlayerInfoZ := 0;
    End;


    result := r;
End;


Function _GetWorldX(MapX, CellX: Integer; SubX: Integer = 0): Integer;
Begin
    result := mapx - 20;
    result := result * 32768 + CellX * 128 + subX * 16 + WorldAddX;
End;

Function _GetWorldY(MapY, CellY: Integer; SubY: Integer = 0): Integer;
Begin
    result := mapy - 18;
    result := result * 32768 + CellY * 128 + subY * 16 + WorldAddY;
End;

Procedure _WorldToXY(wx, wy: Integer; Var CellX, CellY: Integer);
Var
    x, y, sx, sy         : Integer;
Begin
    CellX := -1;
    CellY := -1;
    x := wx - (GeoHead.MX - 20) * 32768;
    y := wy - (GeoHead.MY - 18) * 32768;
    sx := (x Mod 128) Div 16;
    sy := (y Mod 128) Div 16;
    x := x Div 128;
    y := y Div 128;
    x := x * 8 + sx;
    y := y * 8 + sy;
    If x <= 0 Then exit;
    If y <= 0 Then exit;
    If x >= 2048 Then exit;
    If y >= 2048 Then exit;
    CellX := x;
    CellY := y;
End;

Function GeoErrorToStr(err: TGeoError): String;
Begin
    Case err Of
        geOk: result := 'No Errors';
        geFileNotFound: result := 'File Not Found';
        geInvalidFormat: result := 'Invalid Geo Format';
        geTooManyLayers: result := 'Too Many Layers.';
    Else
        result := 'Unknown';
    End;
End;

Procedure ClearLayers;
var
  i: integer;
Begin
    GeoReady := False;
    _LastLayer := 0;
    for i:=0 to Length(GeoLayer)-1 do
      FillChar(GeoLayer[i], sizeof(GeoLayer[i]), $FF);
    //FillChar(CellTypes, sizeof(CellTypes), 0);
End;


Function LoadGeoMap(FName: String): TGeoError;
Var
    f                    : File;
    ID                   : Word;
    x, y, sx, sy, k      : Integer;
    z1                   : SmallInt;
    z2                   : SmallInt;
    num                  : Word;
Begin
    ClearLayers;

    If Not FileExists(FName) Then
    Begin
        result := geFileNotFound;
        Exit;
    End;

    FileMode := 0;
    AssignFile(f, FName);
    Reset(f, 1);

    //read header
    BlockRead(f, GeoHead, sizeof(GeoHead));

    Progress.Caption := 'Importing from conv...';
    Progress.Max := 255;
    Progress.Position := 0;
    Progress.Show;

    For y := 0 To 255 Do
    Begin
        Progress.Position := y;
        Application.ProcessMessages;
        For x := 0 To 255 Do
        Begin
            //read ID
            BlockRead(f, ID, 2);
            If ID = $0000 Then
            Begin
                BlockRead(f, z1, 2);
                BlockRead(f, z2, 2);
                For sy := 0 To 7 Do
                    For sx := 0 To 7 Do
                    Begin
                        GeoLayer[0][x * 8 + sx, y * 8 + sy].m := $F;
                        //SmallInt(z1 and $FFF0);
                        GeoLayer[0][x * 8 + sx, y * 8 + sy].z := SmallInt(z1 And
                            $FFF0);
                        //CellTypes[x, y] := ct0000;
                    End;
            End
            Else If (ID = $0040) Then
            Begin
                //read Z-values
                For sy := 0 To 7 Do
                    For sx := 0 To 7 Do
                    Begin
                        BlockRead(f, z1, 2);
                        GeoLayer[0][x * 8 + sx, y * 8 + sy].m := z1 And $F;
                        GeoLayer[0][x * 8 + sx, y * 8 + sy].z := (SmallInt(z1
                            And
                            $FFF0) Div 2);
                        //CellTypes[x, y] := ct4000;
                    End;
            End
            Else If (ID > $0040) Then
            Begin
                //read Z-values
                For sy := 0 To 7 Do
                    For sx := 0 To 7 Do
                    Begin
                        BlockRead(f, Num, 2);
                        BlockRead(f, z1, 2);
                        GeoLayer[0][x * 8 + sx, y * 8 + sy].m := z1 And $F;
                        GeoLayer[0][x * 8 + sx, y * 8 + sy].z := (smallint(z1
                            And $FFF0) Div 2);

                        If Num > __MaxLayer Then
                        Begin
                            ShowMessage('File Pos: 0x' + inttohex(FilePos(f),
                                8));
                            CloseFile(f);
                            ClearLayers;
                            result := geTooManyLayers;
                            Exit;
                        End;

                        If _LastLayer < Num - 1 Then _LastLayer := Num - 1;

                        For k := 1 To Num - 1 Do
                        Begin
                            BlockRead(f, z2, 2);
                            GeoLayer[k][x * 8 + sx, y * 8 + sy].z := (SmallInt(z2
                                And $FFF0) Div 2);
                            GeoLayer[k][x * 8 + sx, y * 8 + sy].m := z2 And $F;
                        End;

                        //CellTypes[x, y] := ct4100;
                    End;
            End
            Else
            Begin
                CloseFile(f);
                result := geInvalidFormat;
                Exit;
            End;
        End;
    End;
    CloseFile(f);
    result := geOK;
    GeoReady := True;
    Progress.Hide;
End;


Function LoadGeoMap256(MapX, MapY: Integer; b256: TBitmap32; minz, maxz,
    difz: smallint):
    TGeoError;
Var
    b2048                : TBitmap32;
    x, y, z, m           : Integer;
Begin
    ClearLayers;
    GeoHead.MX := MapX;
    GeoHead.MY := MapY;
    b2048 := TBitmap32.Create;
    b2048.Assign(b256);
    b2048.SetSize(2048, 2048);
    b2048.Draw(b2048.BoundsRect, b256.BoundsRect, b256);

    For y := 0 To 2047 Do
        For x := 0 To 2047 Do
        Begin
            z := RedComponent(b2048.Pixel[x, y]);
            z := minz + round(z * (maxz - minz) / 256);
            GeoLayer[0][y, x].z := smallint(z);
            GeoLayer[0][y, x].m := $F;
        End;

    //tune up movement restrictions
    For y := 0 To 2047 Do
        For x := 0 To 2047 Do
        Begin
            If (x > 0) And (x < 2047) And (y > 0) And (y < 2047) Then
            Begin
                If GeoLayer[0][y + 1, x].z - GeoLayer[0][y, x].z > difz Then
                    GeoLayer[0][y, x].m := GeoLayer[0][y, x].m And 11;
                If GeoLayer[0][y - 1, x].z - GeoLayer[0][y, x].z > difz Then
                    GeoLayer[0][y, x].m := GeoLayer[0][y, x].m And 7;
                If GeoLayer[0][y, x + 1].z - GeoLayer[0][y, x].z > difz Then
                    GeoLayer[0][y, x].m := GeoLayer[0][y, x].m And 14;
                If GeoLayer[0][y, x - 1].z - GeoLayer[0][y, x].z > difz Then
                    GeoLayer[0][y, x].m := GeoLayer[0][y, x].m And 13;
            End;
        End;

    b2048.Free;
    GeoReady := True;
    _LastLayer := 0;
    result := geOK;
End;


Procedure SaveGeoMap(FName: String);
Var
    x, y, sx, sy, l, i   : Integer;
    z                    : SmallInt;
    solid                : boolean;
    multi                : boolean;
    f                    : File;
    w                    : word;
Begin
    _Normalize0;
    FileMode := 2;
    assignfile(f, FName);
    rewrite(f, 1);

    GeoHead.w0080 := $80;
    GeoHead.w0010 := $10;
    GeoHead.dw1 := 0;
    GeoHead.dw2 := 0;
    GeoHead.dw3 := 0;

    BlockWrite(f, GeoHead, sizeof(GeoHead));

    For y := 0 To 255 Do
    Begin
        For x := 0 To 255 Do
        Begin
            solid := True;
            multi := False;

            //check for multilayers
            For sy := 0 To 7 Do
            Begin
                For sx := 0 To 7 Do
                Begin
                    //multilayers
                    l := MaxPtLayer(x * 8 + sx, y * 8 + sy);
                    If l > 0 Then
                    Begin
                        Solid := False;
                        Multi := True;
                        Break;
                    End;
                End
            End;

            //check for movement restrictions, difz
            If Solid Then
            Begin
                z := GeoLayer[0][x * 8, y * 8].z;
                For sy := 0 To 7 Do
                Begin
                    For sx := 0 To 7 Do
                    Begin
                        //movement restrictions
                        If GeoLayer[0][x * 8 + sx, y * 8 + sy].m And $F <> $F
                            Then
                        Begin
                            solid := False;
                            Break;
                        End;
                        //difz
                        If abs(GeoLayer[0][x * 8 + sx, y * 8 + sy].z - z) > 30
                            Then
                        Begin
                            solid := False;
                            Break;
                        End;
                    End;
                    If Not Solid Then Break;
                End;
            End;

            If solid Then
            Begin
                //solid cell - get top Z
                z := -32000;
                For sy := 0 To 7 Do
                Begin
                    For sx := 0 To 7 Do
                    Begin
                        If GeoLayer[0][x * 8 + sx, y * 8 + sy].z > z Then
                            z := GeoLayer[0][x * 8 + sx, y * 8 + sy].z;
                    End;
                End;
                //write
                z := smallint(z Or $F);
                w := 0;
                BlockWrite(f, w, 2);
                BlockWrite(f, z, 2);
                BlockWrite(f, z, 2);
                inc(GeoHead.dw3);
                inc(GeoHead.dw2);
            End
            Else
            Begin
                If Not Multi Then
                Begin
                    //cells 8x8 - write each cell
                    w := $40;
                    BlockWrite(f, w, 2);
                    inc(GeoHead.dw2);
                    For sy := 0 To 7 Do
                    Begin
                        For sx := 0 To 7 Do
                        Begin
                            z := GeoLayer[0][x * 8 + sx, y * 8 + sy].z * 2;
                            z := SmallInt(z And $FFF0 Or (GeoLayer[0][x * 8 +
                                sx, y * 8
                                    + sy].m And $F));
                            BlockWrite(f, z, 2);
                            inc(GeoHead.dw1);
                        End;
                    End;
                End
                Else
                Begin
                    //generating header 41+
                    w := 0;
                    For sy := 0 To 7 Do
                    Begin
                        For sx := 0 To 7 Do
                        Begin
                            l := MaxPtLayer(x * 8 + sx, y * 8 + sy) + 1;
                            w := w + l;
                        End;
                    End;

                    If w = $40 Then ShowMessage('DAMN');

                    BlockWrite(f, w, 2);

                    For sy := 0 To 7 Do
                        For sx := 0 To 7 Do
                        Begin
                            l := MaxPtLayer(x * 8 + sx, y * 8 + sy);
                            //write layers
                            w := l + 1;
                            BlockWrite(f, w, 2); //num
                            l := 0;
                            For i := 0 To _LastLayer Do
                            Begin
                                If GeoLayer[i][x * 8 + sx, y * 8 + sy].m <> $FF
                                    Then
                                Begin
                                    z := GeoLayer[i]
                                        [x * 8 + sx, y * 8 + sy].z * 2;
                                    z := SmallInt(z And $FFF0 Or
                                        (GeoLayer[i][x * 8 + sx, y * 8 + sy].m
                                        And $F));
                                    BlockWrite(f, z, 2);
                                    inc(GeoHead.dw1);
                                End;
                            End;
                        End;

                End;

            End;
        End;
    End;

    Seek(f, 0);
    BlockWrite(f, GeoHead, sizeof(GeoHead));

    CloseFile(f);
End;

Procedure SaveGeoMapTurbo(FName: String);
Var
    maxl, lastz, x, y, sx, sy, l, i, k, j: Integer;
    z                    : SmallInt;
    solid                : boolean;
    multi                : boolean;
    //f                    : File;
    //fs: TFileStream;
    ms                   : TMemoryStream;
    w                    : word;
    //tmp                  : Array {[0.._MaxLayer]} Of SmallInt;
    _tmp: array[0..255] of SmallInt;

Begin
    //SetLength(tmp,__MaxLayer);
    _Normalize0;
    //    FileMode := 2;
    //    assignfile(f, FName);
    //    rewrite(f, 1);
    ms := TMemoryStream.Create;

    GeoHead.w0080 := $80;
    GeoHead.w0010 := $10;
    GeoHead.dw1 := 0;
    GeoHead.dw2 := 0;
    GeoHead.dw3 := 0;

    ms.Write(GeoHead, sizeof(GeoHead));


    Progress.Caption := 'Exporting to conv...';
    Progress.Max := 255;
    Progress.Position := 0;
    Progress.Show;

    For y := 0 To 255 Do
    Begin
        Progress.Position := y;
        Application.ProcessMessages;
        For x := 0 To 255 Do
        Begin
            solid := True;
            multi := False;

            //check for multilayers
            For sy := 0 To 7 Do
            Begin
                For sx := 0 To 7 Do
                Begin
                    //multilayers
                    l := MaxPtLayer(x * 8 + sx, y * 8 + sy);
                    If l > 0 Then
                    Begin
                        Solid := False;
                        Multi := True;
                        Break;
                    End;
                End
            End;

            //check for movement restrictions, difz
            If Solid Then
            Begin
                z := GeoLayer[0][x * 8, y * 8].z;
                For sy := 0 To 7 Do
                Begin
                    For sx := 0 To 7 Do
                    Begin
                        //movement restrictions
                        If GeoLayer[0][x * 8 + sx, y * 8 + sy].m And $F <> $F
                            Then
                        Begin
                            solid := False;
                            Break;
                        End;
                        //difz
                        If abs(GeoLayer[0][x * 8 + sx, y * 8 + sy].z - z) > 30
                            Then
                        Begin
                            solid := False;
                            Break;
                        End;
                    End;
                    If Not Solid Then Break;
                End;
            End;

            If solid Then
            Begin
                //solid cell - get top Z
                z := -32000;
                For sy := 0 To 7 Do
                Begin
                    For sx := 0 To 7 Do
                    Begin
                        If GeoLayer[0][x * 8 + sx, y * 8 + sy].z > z Then
                            z := GeoLayer[0][x * 8 + sx, y * 8 + sy].z;
                    End;
                End;
                //write
                z := smallint(z Or $F);
                w := 0;
                ms.Write(w, 2);
                ms.Write(z, 2);
                ms.Write(z, 2);
                inc(GeoHead.dw3);
                inc(GeoHead.dw2);
            End
            Else
            Begin
                If Not Multi Then
                Begin
                    //cells 8x8 - write each cell
                    w := $40;
                    ms.Write(w, 2);
                    inc(GeoHead.dw2);
                    For sy := 0 To 7 Do
                    Begin
                        For sx := 0 To 7 Do
                        Begin
                            z := GeoLayer[0][x * 8 + sx, y * 8 + sy].z * 2;
                            z := SmallInt(z And $FFF0 Or (GeoLayer[0][x * 8 +
                                sx, y * 8
                                    + sy].m And $F));
                            ms.Write(z, 2);
                            inc(GeoHead.dw1);
                        End;
                    End;
                End
                Else
                Begin
                    //generating header 41+
                    w := 0;
                    For sy := 0 To 7 Do
                    Begin
                        For sx := 0 To 7 Do
                        Begin
                            l := MaxPtLayer(x * 8 + sx, y * 8 + sy) + 1;
                            w := w + l;
                        End;
                    End;

                    If w = $40 Then ShowMessage('DAMN');

                    ms.Write(w, 2);

                    For sy := 0 To 7 Do
                        For sx := 0 To 7 Do
                        Begin
                            l := MaxPtLayer(x * 8 + sx, y * 8 + sy);
                            //write layers
                            w := l + 1;
                            ms.Write(w, 2); //num
                            l := 0;
                            For i := 0 To _LastLayer Do
                            Begin
                                If GeoLayer[i][x * 8 + sx, y * 8 + sy].m <> $FF
                                    Then
                                Begin
                                    z := GeoLayer[i]
                                        [x * 8 + sx, y * 8 + sy].z * 2;
                                    z := SmallInt(z And $FFF0 Or
                                        (GeoLayer[i][x * 8 + sx, y * 8 + sy].m
                                        And $F));

                                    _tmp[l] := z;
                                    inc(l);
                                    //ms.Write(z, 2);
                                    inc(GeoHead.dw1);
                                End;
                            End;
                            ArraySort.SmallInt_SelSort(_tmp, 0, l - 1);
                            ms.Write(_tmp, l * 2);
                        End;
                End;
            End;
        End;
    End;

    ms.Seek(0, soFromBeginning);
    ms.Write(GeoHead, sizeof(GeoHead));

    ms.Seek(0, soFromBeginning);
    ms.SaveToFile(FName);
    ms.Free;
    Progress.Hide;
End;


{Function LoadGeoMap09(FName: String): TGeoError;
Var
    f                    : File;
    ID                   : Word;
    x, y, sx, sy, k      : Integer;
    z1                   : SmallInt;
    z2                   : SmallInt;
    num                  : Word;
Begin
    ClearLayers;

    If Not FileExists(FName) Then
    Begin
        result := geFileNotFound;
        Exit;
    End;

    FileMode := 0;
    AssignFile(f, FName);
    Reset(f, 1);

    //read header
    BlockRead(f, GeoHead, sizeof(GeoHead));

    z1:=0;
    z2:=0;

    For y := 0 To 255 Do
    Begin
        For x := 0 To 255 Do
        Begin
            //read ID
            BlockRead(f, ID, 2);
            If ID = $0000 Then
            Begin
                BlockRead(f, z1, 2);
                BlockRead(f, z2, 2);
                For sy := 0 To 7 Do
                    For sx := 0 To 7 Do
                    Begin
                        GeoLayer[0][x * 8 + sx, y * 8 + sy].m := $F;
                        //SmallInt(z1 and $FFF0);
                        GeoLayer[0][x * 8 + sx, y * 8 + sy].z := SmallInt(z1 And
                            $FFF0);
                        CellTypes[x, y] := ct0000;
                    End;
            End
            Else If (ID = $0040) Then
            Begin
                //read Z-values
                For sy := 0 To 7 Do
                    For sx := 0 To 7 Do
                    Begin
                        BlockRead(f, z1, 2);
                        GeoLayer[0][x * 8 + sx, y * 8 + sy].m := z1 And $F;
                        GeoLayer[0][x * 8 + sx, y * 8 + sy].z := (SmallInt(z1
                            And
                            $FFF0) Div 2);
                        CellTypes[x, y] := ct4000;
                    End;
            End
            Else If (ID > $0040) Then
            Begin
                //read Z-values
                For sy := 0 To 7 Do
                    For sx := 0 To 7 Do
                    Begin
                        BlockRead(f, Num, 2);
                        BlockRead(f, z1, 2);
                        GeoLayer[0][x * 8 + sx, y * 8 + sy].m := z1 And $F;
                        GeoLayer[0][x * 8 + sx, y * 8 + sy].z := (smallint(z1
                            And $FFF0) Div 2);

                        If Num > _MaxLayer Then
                        Begin
                            ShowMessage('File Pos: 0x' + inttohex(FilePos(f),
                                8));
                            CloseFile(f);
                            ClearLayers;
                            result := geTooManyLayers;
                            Exit;
                        End;

                        If _LastLayer < Num - 1 Then _LastLayer := Num - 1;

                        For k := 1 To Num - 1 Do
                        Begin
                            BlockRead(f, z2, 2);
                            GeoLayer[k][x * 8 + sx, y * 8 + sy].z := (SmallInt(z2
                                And $FFF0) Div 2);
                            GeoLayer[k][x * 8 + sx, y * 8 + sy].m := z2 And $F;
                        End;

                        CellTypes[x, y] := ct4100;
                    End;
            End
            Else If (ID = $0009) Then
            Begin

                //read Z-values
                For sy := 0 To 7 Do
                    For sx := 0 To 7 Do
                    Begin
                        BlockRead(f, Num, 2);
                        If Num > 0 Then
                        Begin
                            BlockRead(f, z1, 2);
                        End
                        Else
                        Begin
                          z1 := 0;
                        End;
                        GeoLayer[0][x * 8 + sx, y * 8 + sy].m := z1 And $F;
                        GeoLayer[0][x * 8 + sx, y * 8 + sy].z := (smallint(z1
                            And $FFF0) Div 2);

                        If Num > _MaxLayer Then
                        Begin
                            ShowMessage(inttostr(Num));
                            ShowMessage('File Pos: 0x' + inttohex(FilePos(f),
                                8));
                            CloseFile(f);
                            ClearLayers;
                            result := geTooManyLayers;
                            Exit;
                        End;

                        If _LastLayer < Num - 1 Then _LastLayer := Num - 1;

                        For k := 1 To Num - 1 Do
                        Begin
                            BlockRead(f, z2, 2);
                            GeoLayer[k][x * 8 + sx, y * 8 + sy].z := (SmallInt(z2
                                And $FFF0) Div 2);
                            GeoLayer[k][x * 8 + sx, y * 8 + sy].m := z2 And $F;
                        End;

                        CellTypes[x, y] := ct4100;
                    End;

            End
            Else
            Begin
                CloseFile(f);
                result := geInvalidFormat;
                Exit;
            End;
        End;
    End;
    CloseFile(f);
    result := geOK;
    GeoReady := True;
End;}


End.

