unit Sub1;

interface

uses
 Windows,Types,DirectXGraphics, DXGUtils, ut_packages, SysUtils, CSG;

type
  TCUSTOMVERTEX = packed record
    X, Y, Z : Single;
    normVector : TD3DVector;
    Color : DWORD;
  end;

type
 Node = array of Integer;
 Wnode = array of byte;



type
 Ray = packed record
  St,En: TD3DVector;
 end;

type
 TMesh = packed record
  MeshName: String;
  VertexCount, IndexCount: Integer;
  VertexBuf: array of TCustomVertex;
  IndexBuf: array of Word;
  MaxX,MaxY,MaxZ,MinX,MinY,MinZ: single;
 end;

type
 TActor = packed record
  MeshIndex : Integer;
  LocX,LocY,LocZ : single;
  RotP, RotR, RotY : Integer;
  ScaleX,ScaleY,ScaleZ : single;
  MTrform: TD3DMatrix;
  MRotX,MRotY,MRotZ,MScale,Mloc: TD3DMatrix;
  MaxX,MaxY,MaxZ,MinX,MinY,MinZ: single;
 end;
 LandParam = packed record
  h : Single;
 end;
 TPolis = packed record
  PolysName: String;
  PolysCount: Integer;
  Polys: array of TUT_Struct_Polygon;
 end;
 TModel = packed record
  ModelName: String;
  PolysIndex: Integer;
 end;



const
 EPSILON = 0.000001;
 NORD = 8;
 SOUTH = 4;
 WEST = 2;
 EAST = 1;
 

function ReadIdx(fin: Integer) :Integer;
function ReadPrp(fin: Integer) : Integer;
procedure SortActors(var Actors: array of TActor;MeshesCount,ActorsCount: Integer);
function RayIntersect(R : Ray; T: Triangle; var I: TD3DVector): Integer;

implementation

function ReadIdx(fin : Integer) : Integer;
var
b0,b1,b2,b3,b4: byte;
begin
result := 0;
  FileRead(fin,b0,1);
  if (b0 and $40) <> 0 then
    begin
      FileRead(fin,b1,1);
      if (b1 and $80) <> 0 then
        begin
          FileRead(fin,b2,1);
          if (b2 and $80) <> 0 then
            begin
              FileRead(fin,b3,1);
              if (b3 and $80) <> 0 then
                begin
                  FileRead(fin,b4,1);
                  result := b4;
                end;
              result := (result shl 7) or (b3 and $7F);
            end;
          result := (result shl 7) or (b2 and $7F);
        end;
      result := (result shl 7) or (b1 and $7F);
    end;
  result := (result shl 6) or (b0 and $3F);
  if (b0 and $80) <> 0 then
    result := -result;
end;

function ReadPrp(fin: Integer) : Integer;
var
  b: integer;
  infobyte, info_type, info_size, index_b, index_c, index_d, index_e: byte;
  size: integer;
  info_array: boolean;
  buffer: array of byte;
begin
   setlength(buffer, 1024);
   b := ReadIdx(fin);
       if b <> $00 then
          begin
         result:=b;
          FileRead(fin,infobyte,1);
          info_type := infobyte and $0F;
          info_size := (infobyte shr 4) and 7;
          info_array := ((infobyte and $80) <> 0);
          if info_type = otStruct then ReadIdx(fin);
          case info_size of
            0: size := 1;
            1: size := 2;
            2: size := 4;
            3: size := 12;
            4: size := 16;
            5: FileRead(fin,size,1);
            6: FileRead(fin,size,2);
            7: FileRead(fin,size,4);
          else
            size := 0;
          end;

   if info_array and (info_type <> otBool) then
            begin
              FileRead(fin,index_b,1);
              if (index_b and $80) = 0 then
               else if (index_b and $C0) = $80 then
                begin
                  FileRead(fin,index_c,1);
                 end
              else
                begin
                  FileRead(fin,index_c,1);
                  FileRead(fin,index_d,1);
                  FileRead(fin,index_e,1);
                 end;
            end;
        case info_type of
            otBool: ;
            otByte,otInt,otFloat,otObject,otClass,otName,otString,otStr,otArray,otStruct,
            otVector,otRotator,otMap,otFixedArray:
              begin
                setlength(buffer, size);
                FileRead(fin,buffer[0], size);
              end
          else
            begin
              if size > length(buffer) then
                FileRead(fin,buffer[0], length(buffer))
              else
                FileRead(fin,buffer[0], size);
            end;
          end;
        end
      else
      result := $00;
end;

procedure SortActors(var Actors: array of TActor;MeshesCount,ActorsCount: Integer);
var
act,cura,curmi: Integer;
tmpa: array [1..2] of TActor;
begin
cura:=0;
for curmi:=0 to MeshesCount-1 do
for act:=cura to ActorsCount-1 do
 begin
  if Actors[act].MeshIndex=curmi then begin
   CopyMemory(@tmpa[1],@Actors[cura],SizeOf(TActor));
   CopyMemory(@Actors[cura],@Actors[act],SizeOf(TActor));
   CopyMemory(@Actors[act],@tmpa[1],SizeOf(TActor));
   cura:=cura+1;
  end;
 end;
end;

//    Input:  a ray R, and a triangle T
//    Output: *I = intersection point (when it exists)
//    Return: -1 = triangle is degenerate (a segment or point)
//             0 = disjoint (no intersect)
//             1 = intersect in unique point I1
//             2 = are in the same plane


function RayIntersect(R : Ray; T: Triangle; var I: TD3DVector): Integer;
var
u,v,n : TD3DVector;
dir, w0,w : TD3DVector;
r1,a,b : single;
uu, uv, vv, wu, wv, D: single;
s, t1 : single;
begin
u := VectorSub(T.B,T.A);
v := VectorSub(T.C,T.A);
n:=VectorCrossProduct(u,v);
if (n.X=0) and (n.Y=0) and (n.Z=0) then begin result:=-1; exit; end;
dir:=VectorSub(R.En,R.St);
w0 := VectorSub(R.St,T.A);
a := -VectorDotProduct(n,w0);
b := VectorDotProduct(n,dir);
if (b>-EPSILON) and (b<EPSILON) then     // ray is parallel to triangle plane
  if (a = 0) then begin result:=2; exit; end // ray lies in triangle plane
   else begin result:=0; exit; end;              // ray disjoint from plane
r1 := a / b;
if (r1 < 0.0) or (r1> 1.0) then begin result:=0; exit; end;                   // ray goes away from triangle
 I := VectorAdd(R.St,VectorMulS(dir,r1));
 uu := VectorDotProduct(u,u);
 uv := VectorDotProduct(u,v);
 vv := VectorDotProduct(v,v);
 w  := VectorSub(I,T.A);
 wu := VectorDotProduct(w,u);
 wv := VectorDotProduct(w,v);
 D  := (uv * uv) - (uu * vv);
 if D= 0 then begin result:=-1; exit; end;
 s  := ((uv * wv) - (vv * wu)) / D;
 if (s < 0.0) or (s > 1.0) then begin result:=0; exit; end;    // I is outside T
 t1  := ((uv * wu) - (uu * wv)) / D;
 if (t1 < 0.0) or ((s + t1) > 1.0) then begin result:=0; exit; end;  // I is outside T
 result:=1;
end;
end.
